from abc import ABC, abstractmethod
from cachelm.types.chat_history import Message
from cachelm.vectorizers.vectorizer import Vectorizer


class Database(ABC):
    """Abstract base class for a database."""

    def __init__(self, vectorizer: Vectorizer, unique_id: str = "cachelm"):
        self.vectorizer = vectorizer
        self.unique_id = unique_id

    @abstractmethod
    def connect(self) -> bool:
        """Connect to the database."""
        raise NotImplementedError("Subclasses must implement this method.")

    @abstractmethod
    def reset(self):
        """Reset the database."""
        raise NotImplementedError("Subclasses must implement this method.")

    @abstractmethod
    def disconnect(self):
        """Disconnect from the database."""
        raise NotImplementedError("Subclasses must implement this method.")

    @abstractmethod
    def write(self, history: list[Message], response: Message):
        """Write data to the database."""
        raise NotImplementedError("Subclasses must implement this method.")

    @abstractmethod
    def find(self, history: list[Message], distance_threshold=0.1) -> Message | None:
        """Find data in the database."""
        raise NotImplementedError("Subclasses must implement this method.")



from loguru import logger

from cachelm.types.chat_history import Message  # Updated import
from cachelm.databases.database import Database
from cachelm.vectorizers.vectorizer import Vectorizer

try:
    from redisvl.extensions.cache.llm import SemanticCache
    from redisvl.utils.vectorize import CustomTextVectorizer
except ImportError:
    raise ImportError(
        "RedisVL library is not installed. Run `pip install redisvl` to install it."
    )


class RedisVLDatabase(Database):
    """
    Redis database for caching.
    """

    def __init__(
        self, host: str, port: int, vectorizer: Vectorizer, unique_id: str = "cachelm"
    ):
        super().__init__(vectorizer, unique_id)
        self.host = host
        self.port = port
        self.cache = None

    def connect(self) -> bool:
        try:
            self.cache = SemanticCache(
                redis_url=f"redis://{self.host}:{self.port}",
                vectorizer=CustomTextVectorizer(
                    embed=self.vectorizer.embed,
                    embed_many=self.vectorizer.embed_many,
                ),
                overwrite=True,
                name=self.unique_id,
            )
            return True
        except Exception as e:
            logger.error(f"Error connecting to Redis: {e}")
            return False

    def disconnect(self):
        if self.cache:
            self.cache.disconnect()

    def reset(self):
        """
        Reset the database.
        """
        try:
            self.cache.clear()
            logger.info("Redis database reset.")
        except Exception as e:
            logger.error(f"Error resetting Redis: {e}")

    def write(self, history: list[Message], response: Message):
        """
        Write data to the Redis database.
        """
        try:
            prompt = "\n".join([msg.to_formatted_str() for msg in history])
            response_str = response.to_json_str()
            logger.info(f"Writing to Redis: {prompt} -> {response_str}")
            self.cache.store(
                prompt=prompt,
                response=response_str,
            )
        except Exception as e:
            logger.error(f"Error writing to Redis: {e}")

    def find(self, history: list[Message], distance_threshold=0.2) -> Message | None:
        """
        Find data in the database.
        """
        try:
            prompt = "\n".join([msg.to_formatted_str() for msg in history])
            res = self.cache.check(
                prompt=prompt,
                distance_threshold=distance_threshold,
            )
            if res is not None and len(res) > 0:
                response_str = res[0].get("response", "")
                logger.info(f"Found in Redis: {response_str[0:50]}...")
                return Message.from_json_str(response_str)
            return None
        except Exception as e:
            logger.error(f"Error finding from redis: {e}")
            return None



from uuid import uuid4
import openai
from openai.types.chat.chat_completion import ChatCompletion, Choice
from openai.types.chat.chat_completion_message import ChatCompletionMessage
import openai.types.chat.chat_completion_chunk as chat_completion_chunk
from typing import Any, Generic, Literal, TypeVar
from cachelm.adaptors.adaptor import Adaptor
from openai import NotGiven
from loguru import logger
from cachelm.types.chat_history import Message, ToolCall  # Use correct import
from openai.types.chat.chat_completion_message_tool_call import (
    ChatCompletionMessageToolCall,
    Function,
)

T = TypeVar("T", openai.OpenAI, openai.AsyncOpenAI)


class OpenAIAdaptor(Adaptor[T], Generic[T]):
    def _preprocess_chat(self, *args, **kwargs) -> ChatCompletion | None:
        """
        Preprocess the chat messages to set the history.
        """
        if kwargs.get("messages") is not None:
            logger.info("Setting history")
            # Convert dicts to Message objects if needed
            messages = [
                (
                    msg
                    if isinstance(msg, Message)
                    else Message(
                        role=msg.get("role", ""),
                        content=msg.get("content", ""),
                        tool_calls=msg.get("tool_calls"),
                    )
                )
                for msg in kwargs["messages"]
            ]
            self.set_history(messages)
        cached = self.get_cache()
        if cached is not None:
            logger.info("Found cached response")
            # cached is a Message object
            res = ChatCompletion(
                id=str(uuid4()),
                choices=[
                    Choice(
                        index=0,
                        finish_reason="stop",
                        message=ChatCompletionMessage(
                            role=cached.role,
                            content=cached.content,
                            tool_calls=(
                                [
                                    ChatCompletionMessageToolCall(
                                        id=str(uuid4()),
                                        function=Function(
                                            name=tool_call.tool,
                                            arguments=tool_call.args,
                                        ),
                                    )
                                    for tool_call in cached.tool_calls
                                ]
                                if cached.tool_calls is not None
                                else None
                            ),
                        ),
                    )
                ],
                created=0,
                model=kwargs["model"],
                object="chat.completion",
            )
            return res
        return None

    def _preprocess_streaming_chat(
        self, *args, **kwargs
    ) -> openai.Stream[chat_completion_chunk.ChatCompletionChunk] | None:
        """
        Preprocess the streaming chat messages to set the history.
        """
        if kwargs.get("messages") is not None:
            logger.info("Setting history")
            messages = [
                (
                    msg
                    if isinstance(msg, Message)
                    else Message(
                        role=msg.get("role", ""),
                        content=msg.get("content", ""),
                        tool_calls=msg.get("tool_calls"),
                    )
                )
                for msg in kwargs["messages"]
            ]
            self.set_history(messages)
        cached = self.get_cache()
        if cached is not None:
            logger.info("Found cached response")

            def cached_iterator():
                yield chat_completion_chunk.ChatCompletionChunk(
                    id=str(uuid4()),
                    choices=[
                        chat_completion_chunk.Choice(
                            index=0,
                            finish_reason="stop",
                            delta=chat_completion_chunk.ChoiceDelta(
                                role=cached.role,
                                content=cached.content,
                                tool_calls=(
                                    [
                                        chat_completion_chunk.ChoiceDeltaToolCall(
                                            id=str(uuid4()),
                                            function=chat_completion_chunk.ChoiceDeltaToolCallFunction(
                                                name=tool_call.tool,
                                                arguments=tool_call.args,
                                            ),
                                        )
                                        for tool_call in cached.tool_calls
                                    ]
                                    if cached.tool_calls is not None
                                    else None
                                ),
                            ),
                        )
                    ],
                    created=0,
                    model=kwargs["model"],
                    object="chat.completion.chunk",
                )

            return cached_iterator()
        return None

    def _preprocess_streaming_chat_async(
        self, *args, **kwargs
    ) -> openai.AsyncStream[chat_completion_chunk.ChatCompletionChunk] | None:
        """
        Preprocess the streaming chat messages to set the history.
        """
        if kwargs.get("messages") is not None:
            logger.info("Setting history")
            messages = [
                (
                    msg
                    if isinstance(msg, Message)
                    else Message(
                        role=msg.get("role", ""),
                        content=msg.get("content", ""),
                        tool_calls=msg.get("tool_calls"),
                    )
                )
                for msg in kwargs["messages"]
            ]
            self.set_history(messages)
        cached = self.get_cache()
        if cached is not None:
            logger.info("Found cached response")

            async def cached_iterator():
                yield chat_completion_chunk.ChatCompletionChunk(
                    id=str(uuid4()),
                    choices=[
                        chat_completion_chunk.Choice(
                            index=0,
                            finish_reason="stop",
                            delta=chat_completion_chunk.ChoiceDelta(
                                role=cached.role,
                                content=cached.content,
                                tool_calls=(
                                    [
                                        chat_completion_chunk.ChoiceDeltaToolCall(
                                            function=chat_completion_chunk.ChoiceDeltaToolCallFunction(
                                                name=tool_call.tool,
                                                arguments=tool_call.args,
                                            )
                                        )
                                        for tool_call in cached.tool_calls
                                    ]
                                    if cached.tool_calls is not None
                                    else None
                                ),
                            ),
                        )
                    ],
                    created=0,
                    model=kwargs["model"],
                    object="chat.completion.chunk",
                )

            return cached_iterator()
        return None

    def _postprocess_chat(self, completion: ChatCompletion) -> None:
        """
        Postprocess the chat messages to set the history.
        """
        msg = completion.choices[0].message
        message_obj = Message(
            role=msg.role,
            content=msg.content,
            tool_calls=(
                [
                    ToolCall(tool_call.function.name, tool_call.function.arguments)
                    for tool_call in msg.tool_calls
                ]
                if msg.tool_calls is not None
                else None
            ),
        )
        self.add_assistant_message(message_obj)

    def _postprocess_streaming_chat(
        self, response: openai.Stream[chat_completion_chunk.ChatCompletionChunk]
    ) -> Any:
        """
        Postprocess the streaming chat messages to set the history.
        """
        full_content = ""
        tool_calls = None
        role = "assistant"
        for chunk in response:
            delta = chunk.choices[0].delta
            if hasattr(delta, "content") and delta.content is not None:
                full_content += delta.content
            if hasattr(delta, "tool_calls") and delta.tool_calls is not None:
                tool_calls = (
                    [
                        ToolCall(tool_call.function.name, tool_call.function.arguments)
                        for tool_call in delta.tool_calls
                    ]
                    if delta.tool_calls is not None
                    else None
                )
            if hasattr(delta, "role") and delta.role is not None:
                role = delta.role
            yield chunk
        message_obj = Message(
            role=role,
            content=full_content,
            tool_calls=tool_calls if tool_calls is not None else None,
        )
        self.add_assistant_message(message_obj)

    async def _postprocess_streaming_chat_async(
        self, response: openai.AsyncStream[chat_completion_chunk.ChatCompletionChunk]
    ) -> Any:
        """
        Postprocess the streaming chat messages to set the history.
        """
        full_content = ""
        tool_calls = None
        role = "assistant"
        async for chunk in response:
            delta = chunk.choices[0].delta
            if hasattr(delta, "content") and delta.content is not None:
                full_content += delta.content
            if hasattr(delta, "tool_calls") and delta.tool_calls is not None:
                tool_calls = (
                    [
                        ToolCall(tool_call.function.name, tool_call.function.arguments)
                        for tool_call in delta.tool_calls
                    ]
                    if delta.tool_calls is not None
                    else None
                )
            if hasattr(delta, "role") and delta.role is not None:
                role = delta.role
            yield chunk
        message_obj = Message(
            role=role,
            content=full_content,
            tool_calls=tool_calls,
        )
        self.add_assistant_message(message_obj)

    def _get_adapted_openai_sync(adaptorSelf, module: openai.OpenAI) -> openai.OpenAI:
        """
        Get the adapted OpenAI API for synchronous calls.
        """
        base = module
        completions = base.chat.completions

        class AdaptedCompletions(completions.__class__):
            def create_with_stream(
                self,
                *args,
                stream: Literal[True],
                **kwargs,
            ):
                logger.info("Creating completion")
                cached = adaptorSelf._preprocess_streaming_chat(
                    *args, stream=stream, **kwargs
                )
                if cached is not None:
                    logger.info("Found cached response")
                    return cached
                parent = super()
                res = parent.create(*args, stream=stream, **kwargs)
                iterator = adaptorSelf._postprocess_streaming_chat(res)
                return iterator

            def create_without_stream(
                self,
                *args,
                stream: Literal[False] | NotGiven | None = NotGiven,
                **kwargs,
            ):
                logger.info("Creating completion")
                cached = adaptorSelf._preprocess_chat(*args, stream=stream, **kwargs)
                if cached is not None:
                    logger.info("Found cached response")
                    return cached
                parent = super()
                res = parent.create(*args, **kwargs)
                adaptorSelf._postprocess_chat(res)
                logger.info("Storing response in cache")
                return res

            def create(
                self,
                *args,
                **kwargs,
            ):
                logger.info(
                    f"Creating completion with streaming = {kwargs.get('stream')}"
                )
                if kwargs.get("stream") is True:
                    return self.create_with_stream(*args, **kwargs)
                else:
                    return self.create_without_stream(*args, **kwargs)

        base.chat.completions = AdaptedCompletions(
            client=base.chat.completions._client,
        )

        return base

    def _get_adapted_openai_async(
        adaptorSelf, module: openai.AsyncOpenAI
    ) -> openai.AsyncOpenAI:
        """
        Get the adapted OpenAI API for asynchronous calls.
        """
        base = module
        completions = base.chat.completions

        class AdaptedCompletions(completions.__class__):
            async def create_with_stream(
                self,
                *args,
                stream: Literal[True],
                **kwargs,
            ):
                logger.info("Creating completion")
                cached = adaptorSelf._preprocess_streaming_chat_async(
                    *args, stream=stream, **kwargs
                )
                if cached is not None:
                    logger.info("Found cached response")
                    return cached
                parent = super()
                res = await parent.create(*args, stream=stream, **kwargs)
                iterator = adaptorSelf._postprocess_streaming_chat_async(res)
                return iterator

            async def create_without_stream(
                self,
                *args,
                stream: Literal[False] | NotGiven | None = NotGiven,
                **kwargs,
            ):
                logger.info("Creating completion")
                cached = adaptorSelf._preprocess_chat(*args, stream=stream, **kwargs)
                if cached is not None:
                    logger.info("Found cached response")
                    return cached
                parent = super()
                res = await parent.create(*args, **kwargs)
                adaptorSelf._postprocess_chat(res)
                logger.info("Storing response in cache")
                return res

            async def create(
                self,
                *args,
                **kwargs,
            ):
                logger.info(
                    f"Creating completion with streaming = {kwargs.get('stream')}"
                )
                if kwargs.get("stream") is True:
                    return await self.create_with_stream(*args, **kwargs)
                else:
                    return await self.create_without_stream(*args, **kwargs)

        base.chat.completions = AdaptedCompletions(
            client=base.chat.completions._client,
        )

        return base

    def get_adapted(self):
        """
        Get the adapted OpenAI API.
        """
        base = self.module

        if isinstance(base, openai.OpenAI):
            return self._get_adapted_openai_sync(base)

        elif isinstance(base, openai.AsyncOpenAI):
            return self._get_adapted_openai_async(base)
        else:
            raise TypeError(
                f"Unsupported OpenAI module type: {type(base)}. "
                "Expected openai.OpenAI or openai.AsyncOpenAI."
            )



from cachelm.vectorizers.vectorizer import Vectorizer
import unittest


class TestVectorizer(unittest.TestCase):
    def _test_helper(self, vectorizer: Vectorizer):
        """
        Test the vectorizer.
        """
        text = "Hello, world!"
        embedding = vectorizer.embed(text)
        assert isinstance(embedding, list), "Embedding should be a list"
        assert len(embedding) > 0, "Embedding should not be empty"
        assert isinstance(embedding[0], float), "Embedding should be a list of floats"

        multiple_texts = ["Hello, world!", "Goodbye, world!"]
        embeddings = vectorizer.embed_many(multiple_texts)
        assert isinstance(embeddings, list), "Embeddings should be a list"
        assert len(embeddings) == len(
            multiple_texts
        ), "Embeddings length should match input texts length"

        for embedding in embeddings:
            assert isinstance(embedding, list), "Each embedding should be a list"
            assert len(embedding) > 0, "Each embedding should not be empty"
            assert isinstance(
                embedding[0], float
            ), "Each embedding should be a list of floats"
            countSame = embeddings.count(embedding)
            assert countSame == 1, "Each embedding should be unique"
            assert len(embedding) == len(
                embeddings[0]
            ), "All embeddings should have the same length"

    def test_fastembed_vectorizer(self):
        """
        Test the FastEmbed vectorizer.
        """
        from cachelm.vectorizers.fastembed import FastEmbedVectorizer

        vectorizer = FastEmbedVectorizer()
        self._test_helper(vectorizer)

    def test_chroma_vectorizer(self):
        """
        Test the Chroma vectorizer.
        """
        from cachelm.vectorizers.chroma import ChromaVectorizer

        vectorizer = ChromaVectorizer()
        self._test_helper(vectorizer)

    def test_redisvl_vectorizer(self):
        """
        Test the RedisVL vectorizer.
        """
        from cachelm.vectorizers.redisvl import RedisvlVectorizer

        vectorizer = RedisvlVectorizer()
        self._test_helper(vectorizer)



from abc import ABC, abstractmethod
from cachelm.types.chat_history import ChatHistory, Message


class Middleware(ABC):
    """Abstract base class for a middleware."""

    @abstractmethod
    def pre_cache_save(self, message: Message, history: ChatHistory) -> Message | None:
        """Pre-cache hook. Modify the history before caching.
        Args:
            message: The message to be cached.
            history: The chat history to be cached.

        Returns:
            The modified message.
            None if you don't want to cache the message.
        """
        ...

    @abstractmethod
    def post_cache_retrieval(self, message: Message, history: ChatHistory) -> Message:
        """Post-cache hook. Just before returning the response.
        Args:
            message: The message to be cached.
            history: The chat history to be cached.

        Returns:
            The modified message.
        """
        ...



from loguru import logger

from cachelm.types.chat_history import Message  # Correct import

try:
    import clickhouse_connect
    from cachelm.databases.database import Database
    from cachelm.vectorizers.vectorizer import Vectorizer
except ImportError:
    raise ImportError(
        "clickhouse-connect library is not installed. Run `pip install clickhouse-connect` to install it."
    )


class ClickHouse(Database):
    """
    ClickHouse database for caching.
    """

    def __init__(
        self,
        host: str,
        port: int,
        user: str,
        password: str,
        vectorizer: Vectorizer,
        database: str = "cachelm",
        unique_id: str = "cachelm",
    ):
        super().__init__(vectorizer, unique_id)
        self.host = host
        self.port = port
        self.user = user
        self.password = password
        self.database = database
        self.client = None
        self.table = f"{self.database}.{self.unique_id}_cache"

    def connect(self) -> bool:
        try:
            self.client = clickhouse_connect.get_client(
                host=self.host,
                port=self.port,
                username=self.user,
                password=self.password,
                database="default",
            )
            self.client.command(f"CREATE DATABASE IF NOT EXISTS {self.database}")
            self.client.command(
                f"""
                CREATE TABLE IF NOT EXISTS {self.table} (
                    id UUID DEFAULT generateUUIDv4(),
                    prompt String,
                    response String,
                    embedding Array(Float32)
                ) ENGINE = MergeTree()
                ORDER BY id
                """
            )
            return True
        except Exception as e:
            logger.error(f"Error connecting to ClickHouse: {e}")
            return False

    def reset(self):
        """
        Reset the ClickHouse database.
        """
        try:
            self.client.command(f"DROP TABLE IF EXISTS {self.table}")
            logger.info("ClickHouse database reset.")
            self.client.command(
                f"""
                CREATE TABLE IF NOT EXISTS {self.table} (
                    id UUID DEFAULT generateUUIDv4(),
                    prompt String,
                    response String,
                    embedding Array(Float32)
                ) ENGINE = MergeTree()
                ORDER BY id
                """
            )
        except Exception as e:
            logger.error(f"Error resetting ClickHouse: {e}")

    def disconnect(self):
        """
        Disconnect from the ClickHouse database.
        """
        self.client = None

    def write(self, history: list[Message], response: Message):
        """
        Write data to the ClickHouse database.
        """
        # Serialize history as a JSON string of message JSONs
        prompt = "\n".join([msg.to_formatted_str() for msg in history])
        response_str = response.to_json_str()
        logger.info(f"Writing to ClickHouse: {prompt} -> {response_str}")
        try:
            # For embedding, you may want to use only the text content
            prompt_text = " ".join([msg.content for msg in history])
            embedding = self.vectorizer.embed(prompt_text)
            self.client.insert(
                self.table,
                [
                    [prompt, response_str, embedding],
                ],
                column_names=["prompt", "response", "embedding"],
            )
        except Exception as e:
            logger.error(f"Error writing to ClickHouse: {e}")

    def find(self, history: list[Message], distance_threshold=0.2) -> Message | None:
        """
        Find data in the ClickHouse database using cosine similarity.
        """
        try:
            prompt_text = "\n".join([msg.to_formatted_str() for msg in history])
            embedding = self.vectorizer.embed(prompt_text)
            logger.debug(f"Finding in ClickHouse: {prompt_text}")
            query = f"""
                SELECT response, 
                    1 - (dotProduct(embedding, %(embedding)s) / (length(embedding) * length(%(embedding)s))) AS similarity
                FROM {self.table}
                ORDER BY similarity DESC
                LIMIT 1
            """
            result = self.client.query(query, parameters={"embedding": embedding})
            if result.result_rows and len(result.result_rows) > 0:
                response_str, similarity = result.result_rows[0]
                if similarity >= (1 - distance_threshold):
                    logger.info(f"Found in ClickHouse: {response_str[0:50]}...")
                    return Message.from_json_str(response_str)
            return None
        except Exception as e:
            logger.error(f"Error finding from ClickHouse: {e}")
            return None



from uuid import uuid4
from cachelm.types.chat_history import Message  # Use the correct import
from cachelm.databases.database import Database
from cachelm.vectorizers.vectorizer import Vectorizer
from loguru import logger

try:
    import chromadb
except ImportError:
    raise ImportError(
        "ChromaDB library is not installed. Run `pip install chromadb` to install it."
    )


class ChromaDatabase(Database):
    """
    Chroma database for caching.
    """

    def __init__(
        self, vectorizer: Vectorizer, persistant=True, unique_id: str = "cachelm"
    ):
        super().__init__(vectorizer, unique_id)
        self.client = None
        self.collection = None
        self.persistant = persistant

    def __get_adapted_embedding_function(self, vectorizer: Vectorizer):
        class AdaptedEmbeddingFunction(chromadb.EmbeddingFunction):
            def __call__(self, input: chromadb.Documents) -> chromadb.Embeddings:
                return vectorizer.embed_many(input)

        return AdaptedEmbeddingFunction()

    def reset(self):
        """
        Reset the database.
        """
        try:
            if self.client:
                self.client.delete_collection(self.unique_id)
                logger.info("Chroma database reset.")
                self.collection = self.client.get_or_create_collection(
                    self.unique_id,
                    embedding_function=self.__get_adapted_embedding_function(
                        self.vectorizer
                    ),
                )
                logger.info("Chroma database reconnected.")
        except Exception as e:
            logger.error(f"Error resetting Chroma: {e}")

    def connect(self) -> bool:
        try:
            self.client = (
                chromadb.Client()
                if not self.persistant
                else chromadb.PersistentClient()
            )
            self.collection = self.client.get_or_create_collection(
                self.unique_id,
                embedding_function=self.__get_adapted_embedding_function(
                    self.vectorizer
                ),
            )
            return True
        except Exception as e:
            logger.error(f"Error connecting to Chroma: {e}")
            return False

    def disconnect(self):
        pass

    def write(self, history: list[Message], response: Message):
        logger.info(f"Writing to Chroma: {history} -> {response}")
        try:
            history_strs = [msg.to_formatted_str() for msg in history]
            self.collection.add(
                ids=[str(uuid4())],
                documents=["\n".join(history_strs)],
                metadatas=[{"response": response.to_json_str()}],
            )
        except Exception as e:
            logger.error(f"Error writing to Chroma: {e}")

    def find(self, history: list[Message], distance_threshold=0.2) -> Message | None:
        try:
            history_strs = [msg.to_formatted_str() for msg in history]
            res = self.collection.query(
                query_texts=["\n".join(history_strs)], n_results=1
            )
            if res is not None and len(res.get("ids", [[]])[0]) > 0:
                distance = res.get("distances", [[1.0]])[0][0]
                logger.info(f"Distance: {distance}")
                if distance > distance_threshold:
                    logger.info(f"Distance too high: {distance} > {distance_threshold}")
                    return
                response_str = res.get("metadatas", [[{}]])[0][0].get("response", None)
                if response_str is None:
                    logger.info("No response found")
                    return
                logger.info(f"Found in Chroma: {response_str[:50]}...")
                return Message.from_json_str(response_str)
            logger.info(f"Found in Chroma: {res}...")
            return
        except Exception as e:
            logger.error(f"Error finding from Chroma: {e}")
            return
